
import json
from typing import Any
import datetime
from django.forms import *
from django.utils.translation import gettext_lazy as _
from django.forms.models import ModelChoiceIterator, ModelMultipleChoiceField
from django.db.models import Model, QuerySet, Manager
from .models import Token, Profile
from django.db import transaction
from django.db.models import Manager
from .exceptions import JsonResponseException
from .utils import absolute_url, build_url
from .serializer import Serializer
from slth import ENDPOINTS


DjangoModelForm = ModelForm
DjangoForm = Form
DjangoModelChoiceField = ModelChoiceField
DjangoMultipleChoiceField = MultipleChoiceField
DjangoModelMultipleChoiceField = ModelMultipleChoiceField
DjangoFileField = FileField
DjangoImageField = ImageField

MASKS = dict(
    cpf_cnpj='999.999.999-99|99.999.999/9999-99',
    cpf='999.999.999-99',
    cnpj='99.999.999/9999-99',
    telefone='(99) 99999-9999',
)

class FormController:

    def __init__(self, form):
        super().__init__()
        self.form = form
        self.controls = dict(hide=[], show=[], set={})
    
    def hide(self, *names):
        self.controls['hide'].extend(names)

    def show(self, *names):
        self.controls['show'].extend(names)

    def set(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, Model):
                v = dict(id=v.id, value=str(v))
            elif isinstance(v, QuerySet) or isinstance(v, Manager):
                v = [dict(id=v.id, value=str(v)) for v in v]
            elif isinstance(v, bool):
                v = str(v).lower()
            elif isinstance(v, datetime.datetime):
                v = v.strftime('%Y-%m-%d %H:%M')
            elif isinstance(v, datetime.date):
                v = v.strftime('%Y-%m-%d')
            self.controls['set'][k] = v

    def get(self, name, value, default=None):
        if value is None:
            return default
        else:
            try:
                value = self.form.fields[name].to_python(value)
            except KeyError:
                pass
            except ValidationError:
                pass
            return default if value is None else value
        
    def clear(self):
        self.controls['show'].clear()
        self.controls['hide'].clear()
        self.controls['set'].clear()

    def on_change(self, field_name):
        self.clear()
        getattr(self.form._endpoint, f'on_{field_name}_change')(self, self.values())
        return self.controls
    
    def get_field_queryset(self, fname, qs):
        method_attr = None
        method_name = f'get_{fname}_queryset'
        if hasattr(self.form._endpoint, method_name):
            method_attr = getattr(self.form._endpoint, method_name)
        elif hasattr(self.form, 'instance') and hasattr(self.form.instance, method_name):
            method_attr = getattr(self.form.instance, method_name)

        if method_attr:
            qs = method_attr(qs, self.values())
        return qs
    
    def values(self):
        data = dict(**self.controls['set'])
        for name in self.form.fields:
            value = self.get(name, self.form.request.GET.get(name))
            if value:
                data[name] = value
        return data

class FormMixin:

    def fieldset(self, title, fields):
        self.fieldsets[title] = fields
        return self

    def parse_json(self):
        content_type = self.request.META.get('CONTENT_TYPE')
        method = self.request.method.lower()
        if content_type and content_type.lower() in 'application/json':
            if method == 'get' or method == 'post':
                d1 = json.loads(self.request.body.decode()) if self.request.body else {}
                d2 = self.request.GET if method == 'get' else self.request.POST
                d2._mutable = True
                d2.clear()
                for k, v in d1.items():
                    if isinstance(v, list):
                        for i, item in enumerate(v):
                            prefix = f'{k}__{i}'
                            d2[f'{prefix}__id'] = item.get('id')
                            for k1, v1 in item.items():
                                d2[f'{prefix}__{k1}'] = v1
                    elif isinstance(v, dict):
                        prefix = k
                        d2[f'{prefix}__id'] = v.get('id')
                        for k1, v1 in v.items():
                            d2[f'{prefix}__{k1}'] = v1
                    else:
                        d2[k] = v
                d2._mutable = False

    @classmethod
    def get_metadata(cls, name, default=None):
        metaclass = getattr(cls, 'Meta', None)
        return getattr(metaclass, name, default) if metaclass else default

    def serialize(self):
        return self.to_dict()
        
    def display(self, serializable):
        self._display = serializable
        return self

    def to_dict(self, prefix=None):
        field_name = self.request.GET.get('on_change')
        if field_name:
            raise JsonResponseException(self.controller.on_change(field_name))
        
        data = dict(
            type='form', key=self._key, method=self._method, title=self.get_metadata('title', self._title), icon=self.get_metadata('icon'),
            style=self.get_metadata('style'), url=absolute_url(self.request), info=self._info
        )
        data.update(controls=self.controller.controls, width=self.get_metadata('width', '100%'))
        if self._display:
            if isinstance(self._display, Serializer):
                # self._display.request = self.request
                data.update(display=self._display.serialize(forward_exception=True)['data'])
        choices_field_name = self.request.GET.get('choices')
        if self.fieldsets:
            fieldsetlist = []
            for title, names in self.fieldsets.items():
                fields = []
                if prefix:
                    value = self.instance.pk if isinstance(self, ModelForm) else ''
                    fields.append([dict(type='hidden', name=f'{prefix}__id', value=value, label='ID')])
                for name in names:
                    if isinstance(name, str):
                        field = self.serialize_field(name, self.fields[name], prefix, choices_field_name)
                        fields.append([field])
                    else:
                        fields.append(
                            [self.serialize_field(name, self.fields[name], prefix, choices_field_name) for name in name]
                        )
                if len(fields)==1 and isinstance(fields[0], list) and fields[0][0]['type'] == 'inline':
                    fieldsetlist.append(fields[0][0])
                else:
                    fieldsetlist.append(dict(type='fieldset', title=title, fields=fields))
            data.update(fieldsets=fieldsetlist)
        else:
            fields = []
            if prefix:
                value = self.instance.pk if isinstance(self, ModelForm) else ''
                fields.append(dict(type='hidden', name=f'{prefix}__id', value=value, label='ID'))
            for name, field in self.fields.items():
                fields.append(self.serialize_field(name, field, prefix, choices_field_name))
            data.update(fields=fields)
        return data
    
    def serialize_field(self, name, field, prefix, choices_field_name):
        if isinstance(field, InlineFormField) or isinstance(field, InlineModelField):
            value = []
            instances = []
            is_one_to_one = field.max == field.min == 1
            if isinstance(self, ModelForm):
                if isinstance(field, OneToOneField):
                    instances.append(getattr(self.instance, name) if self.instance.id else None)
                elif isinstance(field, OneToManyField):
                    instances.extend(getattr(self.instance, name).all() if self.instance.id else ())
            for i, instance in enumerate(instances):
                prefix = name if is_one_to_one else f'{name}__{i}'
                kwargs = dict(instance=instance, request=self.request) if isinstance(field, InlineModelField) else dict(endpoint=self._endpoint)
                value.append(field.form(**kwargs).to_dict(prefix=prefix))
            if not is_one_to_one:
                value.append(field.form(endpoint=self._endpoint).to_dict(prefix=f'{name}__n'))
            required = getattr(field, 'required2', field.required)
            label = field.label.title() if field.label else field.label
            data = dict(type='inline', min=field.min, max=field.max, name=name, count=len(instances), label=label, required=required, value=value)
        else:
            extra = {}
            ftype = FIELD_TYPES.get(type(field).__name__, 'text')
            if name in self._values:
                ftype = 'hidden'
                value = self._values[name]
                value = value.pk if isinstance(value, Model) else value
            else:
                value = field.initial or self.initial.get(name)
                if callable(value):
                    value = value()
                
                if isinstance(field, ModelMultipleChoiceField) or isinstance(field, DjangoModelMultipleChoiceField):
                    value = [dict(id=obj.id, label=str(obj)) for obj in value] if value else []
                elif isinstance(field, MultipleChoiceField) or isinstance(field,DjangoMultipleChoiceField):
                    value = value if value else []
                elif value and (isinstance(field, ModelChoiceField) or isinstance(field, DjangoModelChoiceField)):
                    obj = field.queryset.get(pk=value)
                    value = dict(id=obj.id, label=str(obj))
                elif isinstance(field, DjangoImageField):
                    value = build_url(self.request, value.url) if value else None
                    extra.update(extensions=field.extensions, width=field.width, height=field.height)
                elif isinstance(field, DjangoFileField):
                    value = build_url(self.request, value.url) if value else None
                    extra.update(extensions=field.extensions, max_size=field.max_size)
            
            if isinstance(field.widget, HiddenInput):
                ftype = 'hidden'
                value = value['id'] if isinstance(value, dict) else value

            fname = name if prefix is None else f'{prefix}__{name}'
            data = dict(type=ftype, name=fname, label=field.label, required=field.required, value=value, help_text=field.help_text, mask=None)
            data.update(**extra)
            for word in MASKS:
                if word in name:
                    data.update(mask=MASKS[word])
            for word in ('password', 'senha'):
                if word in name:
                    data.update(type='password')
            if isinstance(field, CharField) or isinstance(field, IntegerField):
                mask = getattr(field, 'mask', None)
                if mask:
                    data.update(mask=mask)
            if isinstance(field, CharField) and isinstance(field.widget, Textarea):
                data.update(type='textarea')
            if ftype == 'decimal':
                data.update(mask='decimal')
            elif ftype == 'choice':
                if name in self.request.GET:
                    data.update(type='hidden', value=self.request.GET[name])
                else:
                    pick = getattr(field, 'pick', False)
                    if isinstance(field.choices, ModelChoiceIterator) and not pick:
                        if choices_field_name == fname:
                            qs = field.choices.queryset
                            qs = self.controller.get_field_queryset(fname, qs)
                            if 'term' in self.request.GET:
                                qs = qs.apply_search(self.request.GET['term'])
                            raise JsonResponseException([dict(id=obj.id, value=str(obj)) for obj in qs[0:10]])
                        else:
                            data['choices'] = absolute_url(self.request, f'choices={fname}')
                    else:
                        data['choices'] = [dict(id=str(k), value=v) for k, v in field.choices]
                    if pick:
                        data.update(pick=True)
                    
        attr_name = f'on_{name}_change'
        if hasattr(self._endpoint, attr_name):
            data['onchange'] = absolute_url(self.request, f'on_change={name}')
        if name in self._actions:
            cls = ENDPOINTS[self._actions[name]]
            if cls.instantiate(self.request, self).check_permission():
                data.update(action=cls.get_api_metadata(self.request, absolute_url(self.request)))
        return data
    
    def submit(self):
        data = {}
        errors = {}
        inline_fields = {name: field for name, field in self.fields.items() if isinstance(field, InlineFormField) or isinstance(field, InlineModelField)}
        self.is_valid()
        errors.update(self.errors)
        for inline_field_name, inline_field in inline_fields.items():
            data[inline_field_name] = []
            for i in range(0, inline_field.max):
                is_one_to_one = inline_field.max == inline_field.min == 1
                prefix = inline_field_name if is_one_to_one else f'{inline_field_name}__{i}'
                inline_form_field_name = f'{prefix}__id'
                if inline_form_field_name in self.data:
                    inline_form_data = {}
                    pk = self.data.get(inline_form_field_name)
                    if pk:
                        pk = int(pk)
                        for name in inline_field.form.base_fields:
                            inline_form_field_name = f'{prefix}__{name}'
                            inline_form_data[name] = self.data.get(inline_form_field_name)
                            if self.request.FILES and inline_form_field_name in self.request.FILES:
                                self.request.FILES._mutable = True
                                self.request.FILES[name] = self.request.FILES[inline_form_field_name]
                                self.request.FILES._mutable = False
                        instance = inline_field.form._meta.model.objects.get(pk=abs(pk)) if pk else None
                        if pk < 0:
                            setattr(instance, 'deleting', True)
                        inline_form = inline_field.form(data=inline_form_data, instance=instance, request=self.request)
                        if inline_form.is_valid():
                            if isinstance(inline_form, DjangoModelForm):
                                data[inline_field_name].append(inline_form.instance)
                            else:
                                data[inline_field_name].append(inline_form.cleaned_data)
                        else:
                            errors.update({f'{prefix}__{name}': error for name, error in inline_form.errors.items()}) 
        if errors:
            raise JsonResponseException(dict(type='error', text='Por favor, corrija os erros.', errors=errors))
        else:
            data.update({field_name: self.cleaned_data.get(field_name) for field_name in self.fields if field_name not in inline_fields})
            with transaction.atomic():
                self.cleaned_data = data
                if isinstance(self, DjangoModelForm):
                    for inline_field_name in inline_fields:
                        for obj in self.cleaned_data[inline_field_name]:
                            obj.save()
                            # set one-to-one
                            if hasattr(self.instance, f'{inline_field_name}_id'):
                                if hasattr(obj, 'deleting'):
                                    setattr(self.instance, inline_field_name, None)
                                else:
                                    setattr(self.instance, inline_field_name, obj)
                    self.save()
                    for inline_field_name in inline_fields:
                        for obj in self.cleaned_data[inline_field_name]:
                            if hasattr(obj, 'deleting'):
                                obj.delete()
                return self.cleaned_data

    def settitle(self, title):
        self._title = title 
        return self
    
    def setvalue(self, **kwargs):
        self._values.update(**kwargs)
        return self
    
    def actions(self, **kwargs):
        self._actions.update(kwargs)
        return self
    

class Form(DjangoForm, FormMixin):
    
    def __init__(self, *args, endpoint=None, **kwargs):
        self._display = []
        self._endpoint = endpoint
        self._info = None
        self._values = {}
        self._title = type(self).__name__
        self._actions = {}
        self._method = 'POST'
        self._key = self._title.lower()

        self.fieldsets = {}
        self.fields = {}
        self.request = endpoint.request
        self.controller = FormController(self)
        self.instance = kwargs.pop('instance', None)

        self.parse_json()
        if 'data' not in kwargs:
            if self.request.method.upper() == 'GET':
                data = self.request.GET or None
            elif self.request.method.upper() == 'POST':
                data = self.request.POST or {}
        else:
            data = kwargs['data']
        kwargs.update(data=data)
        if self.request:
            kwargs.update(files=self.request.FILES or None)
        super().__init__(*args, **kwargs)


class ModelForm(DjangoModelForm, FormMixin):

    def __init__(self, instance=None, endpoint=None, **kwargs):
        self._display = []
        self._endpoint = endpoint
        self._info = None
        self._values = {}
        self._title = type(self).__name__
        self._actions = {}
        self._method = 'POST'
        self._key = self._title.lower()

        self.fieldsets = {}
        self.request = endpoint.request
        self.controller = FormController(self)
        
        self.parse_json()
        if 'data' not in kwargs:
            if self.request.method.upper() == 'GET':
                data = self.request.GET or None
            elif self.request.method.upper() == 'POST':
                data = self.request.POST or {}
            else:
                data = None
        else:
            data = kwargs['data']
        kwargs.update(data=data)
        if self.request:
            kwargs.update(files=self.request.FILES or None)
        super().__init__(instance=instance, **kwargs)

class InlineFormField(Field):
    def __init__(self, *args, form=None, min=1, max=3, **kwargs):
        self.form = form
        self.max = max
        self.min = min
        super().__init__(*args,  **kwargs)
        self.required = False


class InlineModelField(Field):
    def __init__(self, model, *args, fields='__all__', exclude=(), min=1, max=3, **kwargs):
        if fields == '__all__':
            self.form = modelform_factory(model, form=ModelForm, fields=fields, exclude=exclude)
        else:
            field_names = []
            for field_name in fields:
                if isinstance(field_name, str):
                    field_names.append(field_names)
                else:
                    field_names.extend(field_name)
            self.form = modelform_factory(model, form=ModelForm, fields=field_names, exclude=exclude)
            self.form.fieldsets = {None: fields}
        self.max = max
        self.min = min
        super().__init__(*args,  **kwargs)
        self.required = False


class OneToManyField(InlineModelField):
    def __init__(self, model, fields='__all__', exclude=(), min=1, max=3, **kwargs):
        super().__init__(model, fields=fields, exclude=exclude, min=min, max=max)

class OneToOneField(InlineModelField):
    def __init__(self, model, fields='__all__', exclude=(), **kwargs):
        super().__init__(model, fields=fields, exclude=exclude, min=1, max=1)

class DecimalField(DecimalField):
    def to_python(self, value):
        if isinstance(value, str) and ',' in value:
            value = value.replace('.', '').replace(',', '.')
        return super().to_python(value)
    
class ColorField(CharField):
    pass

class FileField(FileField):
    def __init__(self, *args, extensions=('pdf',), max_size=5, **kwargs):
        self.extensions = extensions
        self.max_size = max_size
        super().__init__(*args, **kwargs)

class ImageField(ImageField):
    def __init__(self, *args, extensions=('png', 'jpg', 'jpeg'), width=None, height=None, **kwargs):
        self.extensions = extensions
        if width or height:
            self.width = width or height
            self.height = height or width
        else:
            self.width = self.height = 500
        super().__init__(*args, **kwargs)

class ChoiceField(ChoiceField):
    
    def __init__(self, *args, **kwargs):
        self.pick = kwargs.pop('pick', False)
        super().__init__(*args, **kwargs)
    
class MultipleChoiceField(MultipleChoiceField):
    
    def __init__(self, *args, **kwargs):
        self.pick = kwargs.pop('pick', False)
        super().__init__(*args, **kwargs)

class ModelChoiceField(ModelChoiceField):
    
    def __init__(self, *args, **kwargs):
        self.pick = kwargs.pop('pick', False)
        super().__init__(*args, **kwargs)

class ModelMultipleChoiceField(ModelMultipleChoiceField):
    
    def __init__(self, *args, **kwargs):
        self.pick = kwargs.pop('pick', False)
        super().__init__(*args, **kwargs)
        

FIELD_TYPES = {
    'CharField': 'text',
    'EmailField': 'email',
    'DecimalField': 'decimal',
    'BooleanField': 'boolean',
    'DateTimeField': 'datetime',
    'DateField': 'date',
    'IntegerField': 'number',
    'ChoiceField': 'choice',
    'TypedChoiceField' : 'choice',
    'ModelChoiceField': 'choice',
    'MultipleChoiceField': 'choice',
    'ModelMultipleChoiceField': 'choice',
    'ColorField': 'color',
    'FileField': 'file',
    'ImageField': 'file',
}


from django.template.loader import render_to_string


SUCCESS = 'success'
PRIMARY = 'primary'
WARNING = 'warning'
DANGER = 'danger'


class Image(dict):
    def __init__(self, src, width=None, height=None, round=False):
        if width is None and height is None:
            width = 200
            height = 200
        if width and not height:
            height = width
        if height and not width:
            width = height
        self['type'] = 'image'
        self['src'] = src
        self['width'] = width
        self['height'] = height
        self['round'] = round


class FileLink(dict):
    def __init__(self, url, modal=False, icon=None):
        self['type'] = 'filelink'
        self['url'] = url
        self['modal'] = modal
        self['icon'] = icon


class FileViewer(dict):
    def __init__(self, url):
        self['type'] = 'fileviewer'
        self['url'] = url


class QrCode(dict):
    def __init__(self, text):
        self['type'] = 'qrcode'
        self['text'] = text


class Progress(dict):
    def __init__(self, value, style='primary'):
        self['type'] = 'progress'
        self['value'] = int(value or 0)
        self['style'] = style


class Status(dict):
    def __init__(self, style, label):
        self['type'] = 'status'
        self['style'] = style
        self['label'] = str(label)


class Badge(dict):
    def __init__(self, color, label):
        self['type'] = 'badge'
        self['color'] = color
        self['label'] = str(label)

    
class Shell(dict):
    def __init__(self, output):
        self['type'] = 'shell'
        self['output'] = output


class Indicators(dict):
    def __init__(self, title):
        self['type'] = 'indicators'
        self['title'] = title
        self['items'] = []
        self['actions'] = []

    def append(self, name, value):
        self['items'].append(dict(name=str(name), value=value))

    def action(self, label, url, modal=False):
        self['actions'].append(dict(label=str(label), url=url, modal=modal))


class Boxes(dict):
    def __init__(self, title):
        self['type'] = 'boxes'
        self['title'] = str(title)
        self['items'] = []

    def append(self, icon, label, url, style=None):
        self['items'].append(dict(icon=icon, label=str(label), url=url, style=style))

class Info(dict):
    def __init__(self, title, message):
        self['type'] = 'info'
        self['title'] = title
        self['message'] = message
        self['actions'] = []

    def action(self, label, url, modal=False, icon=None):
        self['actions'].append(dict(label=str(label), url=url, modal=modal, icon=icon))


class Warning(dict):
    def __init__(self, title, message):
        self['type'] = 'warning'
        self['title'] = title
        self['message'] = message
        self['actions'] = []

    def action(self, label, url, modal=False, icon=None):
        self['actions'].append(dict(label=str(label), url=url, modal=modal, icon=icon))


class Table(dict):
    def __init__(self, title, subset=None, pagination=None):
        self['type'] = 'table'
        self['title'] = title
        self['actions'] = []
        self['subsets'] = []
        self['subset'] = subset
        self['filters'] = []
        self['flags'] = []
        self['rows'] = []
        self['pagination'] = {}

    def add_subset(self, name, label, count):
        self['subsets'].append(dict(name=name, label=label, count=count))

    def add_action(self, name, label, icon=None, batch=True):
        self['actions'].append(dict(name=name, label=label, icon=icon, batch=batch))

    def add_flag(self, name, label, checked=False):
        self['flags'].append(dict(name=name, label=label, checked=checked))

    def add_filter(self, ftype, name, label, value, choices=None):
        self['filters'].append(dict(type=ftype, name=name, label=label, value=value, choices=choices))

    def pagination(self, size, page, total, sizes):
        self['pagination'].update(size=size, page=page, total=total, sizes=sizes)

    def add_row(self, row):
        self['rows'].append(row)

    def row(self, value=None, checkable=False, deleted=False):
        self['rows'].append([dict(name='#', value=value, checkable=checkable, deleted=deleted)])

    def cell(self, name, value, style=None, url=None, actions=None):
        self['rows'][-1].append(dict(name=name, value=value, style=style, url=url, actions=actions))


class TemplateContent(dict):
    def __init__(self, name, context):
        self['type'] = 'html'
        self['content'] = render_to_string(name, context)

class Banner(dict):
    def __init__(self, src):
        self['type'] = 'banner'
        self['src'] = src


class Map(dict):
    def __init__(self, latitude, longitude, width='100%', height=400):
        self['type'] = 'map'
        self['latitude'] = str(latitude)
        self['longitude'] = str(longitude)
        self['width'] = width
        self['height'] = height


class Steps(dict):
    def __init__(self, icon=None):
        self['type'] = 'steps'
        self['icon'] = icon
        self['steps'] = []

    def append(self, name, done):
        number = len(self['steps']) + 1
        self['steps'].append(dict(number=number, name=name, done=bool(done)))


class WebConf(dict):
    def __init__(self, caller, receiver):
        self['type'] = 'webconf'
        self['caller'] = caller
        self['receiver'] = receiver

class Navbar(dict):
    def __init__(self, title, subtitle=None, logo=None, user=None):
        self['type'] = 'navbar'
        self['title'] = title
        self['subtitle'] = subtitle
        self['logo'] = logo
        self['user'] = user
        self['usermenu'] = []
        self['adder'] = []
        self['tools'] = []
        self['settings'] = []
        self['actions'] = []

    def add_action(self, entrypoint, name, url, modal=True, icon=None):
        self[entrypoint].append(dict(name=name, url=url, modal=modal, icon=icon))

class Menu(dict):
    def __init__(self, items, user=None, image=None):
        self['type'] = 'menu'
        self['items'] = items
        self['user'] = user
        self['image'] = image

class Footer(dict):
    def __init__(self, version):
        self['type'] = 'navbar'
        self['version'] = version

class Application(dict):
    def __init__(self, icon=None, navbar=None, menu=None, footer=None, oauth=()):
        self['type'] = 'application'
        self['icon'] = icon
        self['navbar'] = navbar
        self['menu'] = menu
        self['footer'] = footer
        self['oauth'] = oauth

class Response(dict):

    def __init__(self, message=None, redirect=None, store=None):
        self['type'] = 'response'
        self['message'] = message
        self['redirect'] = redirect
        self['store'] = store or {}


class IconSet(dict):
    def __init__(self):
        self['type'] = 'iconset'


class Grid(dict):
    def __init__(self):
        self['type'] = 'grid'
        self['items'] = []

    def append(self, component):
        self['items'].append(component)



from django.db.models import Model
from .serializer import Serializer
class FormFactory:
    def __init__(self, instance, endpoint=None, method='POST'):
        self._instance = instance
        self._fieldsets = {}
        self._values = {}
        self._fieldlist = []
        self._display = {}
        self._title = None
        self._info = None
        self._actions = {}
        self._initial = {}
        self._choices = {}
        self._empty = False
        self._method = method
        self._hidden= []

    def _append_field(self, field_name):
        if ':' in field_name:
            field_name, action_name = field_name.split(':')
            self._actions[field_name] = action_name
        self._fieldlist.append(field_name)
        return field_name

    def fields(self, *names, **values) -> 'FormFactory':
        not_str = {name for name in names if not isinstance(name, str)}
        if not_str:
            self.fieldset('Dados Gerais', names)
        else:
            for field_name in names:
                self._append_field(field_name)
        for k in values:
            self._append_field(k)
            self.setvalue(**values)
        self._empty = not self._fieldlist
        return self

    def fieldset(self, title, fields) -> 'FormFactory':
        self._fieldsets[title] = []
        for field in fields:
            if isinstance(field, str):
                self._fieldsets[title].append(self._append_field(field))
            else:
                names = []
                for field_name in field:
                    names.append(self._append_field(field_name))
                self._fieldsets[title].append(names)
        return self
    
    def display(self, title, fields) -> 'FormFactory':
        self._display[title] = fields
        return self
    
    def info(self, message) -> 'FormFactory':
        self._info = message
        return self
    
    def initial(self, **kwargs) -> 'FormFactory':
        self._initial.update(kwargs)
        return self
    
    def choices(self, **kwargs) -> 'FormFactory':
        self._choices.update(kwargs)
        return self
    
    def actions(self, **kwargs) -> 'FormFactory':
        self._actions.update(kwargs)
        return self
    
    def setvalue(self, **kwargs) -> 'FormFactory':
        self._values.update(kwargs)
        return self
    
    def settitle(self, title) -> 'FormFactory':
        self._title = title
        return self
    
    def hidden(self, *names):
        self._hidden.extend(names)
        return self

    def form(self, endpoint):
        from .forms import ModelForm, Form
        
        if isinstance(self._instance, Model):
            fieldlist = [field.name for field in type(self._instance)._meta.get_fields() if field.name in self._fieldlist]
            class Form(ModelForm):
                class Meta:
                    model = type(self._instance)
                    fields = () if self._empty else (fieldlist if self._fieldlist else '__all__')
                
        form = Form(instance=self._instance, endpoint=endpoint, initial=self._initial)
        form._key = endpoint.get_api_name()
        form._title = self._title
        form._method = self._method
        form._info = self._info
        form._actions = self._actions
        for name in self._fieldlist:
            if name not in form.fields:
                form.fields[name] = getattr(endpoint, name)
        for name, queryset in self._choices.items():
            form.fields[name].queryset = queryset
        form.fieldsets = self._fieldsets
        form.setvalue(**self._values)
        if self._display:
            serializer = Serializer(self._instance, request=endpoint.request)
            for title, fields in self._display.items():
                serializer.fieldset(title, fields)
            form._display = serializer
        if self._hidden:
            form.controller.hide(*self._hidden)
        return form


import json
from decimal import Decimal
from django.db.models.aggregates import Count, Sum, Avg
from django.core.exceptions import FieldDoesNotExist


MONTHS = 'JAN', 'FEV', 'MAR', 'ABR', 'MAI', 'JUN', 'JUL', 'AGO', 'SET', 'OUT', 'NOV', 'DEZ'
PIE_CHART = 'pie'
POLAR_CHART = 'polar'
DONUT_CHART = 'donut'
BAR_CHART = 'bar'
STACKED_BAR_CHART = 'stacked_bar'
COLUMN_CHART = 'column'
STACKED_COLUMN_CHART = 'stacked_column'
TREE_MAP_CHART = 'tree_map'
LINE_CHART = 'line'
AREA_CHART = 'area'
PROGRESS_CHART = 'progress'


class Statistics(object):

    def __init__(self, qs, title=None, chart=None):
        self.qs = qs
        self.x = None
        self.y = None
        self.func = None
        self.z = 'id'
        self._xfield = None
        self._yfield = None
        self._xdict = {}
        self._ydict = {}
        self._values_dict = None
        self.title = title
        self.chart = chart

    def _calc(self):
        if '__month' in self.x:
            self._xdict = {i + 1: month for i, month in enumerate(MONTHS)}
        if self.y and '__month' in self.y:
            self._ydict = {i + 1: month for i, month in enumerate(MONTHS)}
        if self._values_dict is None:
            self.calc()

    def _xfield_display_value(self, value):
        if hasattr(self._xfield, 'choices') and self._xfield.choices:
            for choice in self._xfield.choices:
                if choice[0] == value:
                    return choice[1]
        return value

    def _yfield_display_value(self, value):
        if hasattr(self._yfield, 'choices') and self._yfield.choices:
            for choice in self._yfield.choices:
                if choice[0] == value:
                    return choice[1]
        return value

    def _clear(self):
        self._xfield = None
        self._yfield = None
        self._xdict = {}
        self._ydict = {}
        self._values_dict = None

    def calc(self):
        self._values_dict = {}
        if self.y:
            values_list = self.qs.values_list(self.x, self.y).annotate(self.func(self.z))
        else:
            values_list = self.qs.values_list(self.x).annotate(self.func(self.z))

        self._xfield = self.qs.model.get_field(self.x.replace('__year', '').replace('__month', ''))
        if self._xdict == {}:
            xvalues = self.qs.values_list(self.x, flat=True).order_by(self.x).distinct()
            if self._xfield.related_model:
                if hasattr(self._xfield.related_model, 'cor'):
                    self.colors = {pk: color for pk, color in self._xfield.related_model.objects.values_list('id', 'cor')}
                self._xdict = {
                    obj.pk: str(obj) for obj in self._xfield.related_model.objects.filter(pk__in=xvalues)
                }
            else:
                self._xdict = {
                    value: value for value in self.qs.values_list(self.x, flat=True)
                }
            if None in xvalues:
                self._xdict[None] = 'Não-Informado'
        if self.y:
            self._yfield = self.qs.model.get_field(self.y.replace('__year', '').replace('__month', ''))
            yvalues = self.qs.values_list(self.y, flat=True).order_by(self.y).distinct()
            if self._ydict == {}:
                if self._yfield.related_model:
                    self._ydict = {
                        obj.pk: str(obj) for obj in self._yfield.related_model.objects.filter(pk__in=yvalues)
                    }
                else:
                    self._ydict = {
                        value: value for value in yvalues
                    }
            self._values_dict = {(vx, vy): calc for vx, vy, calc in values_list}
            if None in yvalues:
                self._ydict[None] = 'Não-Informado'
        else:
            self._ydict = {}
            self._values_dict = {(vx, None): calc for vx, calc in values_list}

    def debug(self):
        print(json.dumps(self.serialize(), indent=4, ensure_ascii=False))

    def serialize(self):
        tx = []
        ty = []
        series = dict()
        self._calc()

        formatter = {True: 'Sim', False: 'Não', None: ''}

        def format_value(value):
            return float(value) if isinstance(value, Decimal) else value

        if self._ydict:
            for i, (yk, yv) in enumerate(self._ydict.items()):
                data = []
                for j, (xk, xv) in enumerate(self._xdict.items()):
                    data.append([formatter.get(xv, str(self._xfield_display_value(xv))), format_value(self._values_dict.get((xk, yk), 0)), '#000000'])
                series.update(**{formatter.get(yv, str(self._yfield_display_value(yv))): data})
            ty = []
            for k, serie in series.items():
                sy = 0
                ly = len(serie)
                for i, item in enumerate(serie):
                    sy += item[1]
                ty.append(sy)
            tx = [0 for i in range(0, ly)]
            # breakpoint()
            for i in range(0, len(tx)):
                for kj in series:
                    tx[i] += series[kj][i][1]
            if len(tx) == len(ty) == 1:
                tx = ty = []
            elif len(tx) > 1 and len(ty) > 1:
                ty.append(sum(ty))
                tx.append(sum(tx))
            elif len(ty) > 1:
                if len(ty) > 2: tx.append(sum(tx))
                ty = []
            elif len(tx) > 1:
                if len(tx) > 2: ty.append(sum(ty))
                tx = []
        else:
            sx = 0
            data = list()
            for j, (xk, xv) in enumerate(self._xdict.items()):
                data.append([formatter.get(xv, str(self._xfield_display_value(xv))), format_value(self._values_dict.get((xk, None), 0)), '#000000'])
                sx+=data[-1][1]
            if data:
                series['default'] = data
            tx = [sx]
        return dict(type='statistics', title=self.title, series=series['default'] if 'default' in series else series, chart=self.chart)

    def count(self, x, y=None):
        self.func = Count
        self.x = x
        self.y = y
        self._calc()
        return self.serialize()

    def sum(self, z, x=None, y=None):
        self.func = Sum
        self.x = x
        self.y = y
        self._calc()
        return self.serialize()



from uuid import uuid1
from django.db.models import Model as DjangoModel
from django.db.models import *
from django.utils.translation import gettext_lazy as _
from .. import ModelMixin

class CharField(CharField):
    def __init__(self, *args, **kwargs):
        self.mask = kwargs.pop('mask', None)
        kwargs.setdefault('max_length', 255)
        super().__init__(*args, **kwargs)

    def formfield(self, *args, **kwargs):
        field = super().formfield(*args, **kwargs)
        field.mask = self.mask
        return field
    
class IntegerField(IntegerField):
    def __init__(self, *args, **kwargs):
        self.mask = kwargs.pop('mask', None)
        super().__init__(*args, **kwargs)

    def formfield(self, *args, **kwargs):
        field = super().formfield(*args, **kwargs)
        field.mask = self.mask
        return field

class ForeignKey(ForeignKey):
    def __init__(self, to, on_delete=None, **kwargs):
        self.pick = kwargs.pop('pick', False)
        self.addable = kwargs.pop('addable', False)
        super().__init__(to, on_delete or CASCADE, **kwargs)
    
    def formfield(self, *args, **kwargs):
        field = super().formfield(*args, **kwargs)
        return field

class ManyToManyField(ManyToManyField):
    def __init__(self, *args, **kwargs):
        self.pick = kwargs.pop('pick', False)
        self.addable = kwargs.pop('addable', False)
        super().__init__(*args, **kwargs)

    def formfield(self, *args, **kwargs):
        field = super().formfield(*args, **kwargs)
        return field

class OneToManyField(ManyToManyField):
    def __init__(self, *args, **kwargs):
        self.fields = kwargs.pop('fields', '__all__')
        super().__init__(*args, **kwargs)

    def formfield(self, *args, **kwargs):
        from .. import forms
        kwargs.update(form_class=forms.OneToManyField, fields=self.fields)
        field = super().formfield(*args, model=self.related_model, **kwargs)
        field.label = self.verbose_name
        return field


class OneToOneField(OneToOneField):
    def __init__(self, *args, **kwargs):
        self.fields = kwargs.pop('fields', '__all__')
        super().__init__(*args, **kwargs)

    def formfield(self, *args, **kwargs):
        from .. import forms
        kwargs.update(form_class=forms.OneToOneField, fields=self.fields)
        field = super().formfield(*args, model=self.related_model, **kwargs)
        field.label = self.verbose_name
        field.required2 = not self.blank
        return field

class DecimalField(DecimalField):
    def __init__(self, *args, **kwargs):
        kwargs['decimal_places'] = kwargs.pop('decimal_places', 2)
        kwargs['max_digits'] = kwargs.pop('max_digits', 9)
        super().__init__(*args, **kwargs)

    def formfield(self, *args, **kwargs):
        from .. import forms
        kwargs.update(form_class=forms.DecimalField)
        return super().formfield(*args, **kwargs)
    
class ColorField(CharField):

    def formfield(self, *args, **kwargs):
        from .. import forms
        kwargs.update(form_class=forms.ColorField)
        return super().formfield(*args, **kwargs)

class TextField(TextField):
    def __init__(self, *args, **kwargs):
        self.formatted= kwargs.pop('formatted', False)
        super().__init__(*args, **kwargs)

class FileField(FileField):
    def __init__(self, *args, extensions=('pdf',), max_size=5, **kwargs):
        self.extensions= extensions
        self.max_size = max_size
        super().__init__(*args, **kwargs)

    def formfield(self, *args, **kwargs):
        from .. import forms
        kwargs.update(extensions=self.extensions, max_size=self.max_size)
        kwargs.update(form_class=forms.FileField)
        return super().formfield(*args, **kwargs)
    
    def generate_filename(self, instance, filename):
        filename = '{}.{}'.format(uuid1().hex, filename.split('.')[-1].lower())
        print(filename)
        return super().generate_filename(instance, filename)
    
class ImageField(ImageField):
    def __init__(self, *args, extensions=('png', 'jpg', 'jpeg'), width=None, height=None, **kwargs):
        self.extensions= extensions
        self.width = width
        self.height = height
        super().__init__(*args, **kwargs)

    def formfield(self, *args, **kwargs):
        from .. import forms
        kwargs.update(extensions=self.extensions, width=self.width, height=self.height)
        kwargs.update(form_class=forms.ImageField)
        return super().formfield(*args, **kwargs)
    
    def generate_filename(self, instance, filename):
        filename = '{}.{}'.format(uuid1().hex, filename.split('.')[-1].lower())
        print(filename)
        return super().generate_filename(instance, filename)

class Model(DjangoModel, ModelMixin):
    class Meta:
        abstract = True




import json
import types
import inspect
from .models import Token
from django.apps import apps
from typing import TypeVar, Generic
from django.core.cache import cache
from django.conf import settings
from django.utils.text import slugify
from django.db import models
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt
from .factory import FormFactory
from django.core.exceptions import ValidationError
from slth import forms
from django.contrib.auth import authenticate
from .forms import ModelForm, Form
from .serializer import serialize, Serializer
from .components import Application as Application_, Navbar, Menu, Footer, Response, Boxes, IconSet
from .exceptions import JsonResponseException
from .utils import build_url, append_url
from .models import PushSubscription, Profile, User
from slth.queryset import QuerySet
from .notifications import send_push_web_notification
from slth import APPLICATON, ENDPOINTS
from . import oauth


T = TypeVar("T")

class ApiResponse(JsonResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self["Access-Control-Allow-Origin"] = "*"
        self["Access-Control-Allow-Headers"] = "*"
        self["Access-Control-Allow-Methods"] = "GET, POST, OPTIONS, PUT, DELETE, PATCH";
        self["Access-Control-Max-Age"] = "600"


class EnpointMetaclass(type):
    def __new__(mcs, name, bases, attrs):
        cls = super().__new__(mcs, name, bases, attrs)
        bases_names = [cls.__name__ for cls in bases]
        if name not in ('Endpoint', 'ChildEndpoint') and '_ChildEndpoint' not in bases_names:
            ENDPOINTS[cls.__name__.lower()] = cls
        if 'AdminEndpoint' in bases_names[0:1]:
            model = cls.__orig_bases__[0].__args__[0]
            items = (
                ('Cadastrar', AddEndpoint[model], 'plus'),
                ('Editar', EditEndpoint[model], 'pen'),
                ('Visualizar', ViewEndpoint[model], 'eye'),
                ('Excluir', DeleteEndpoint[model], 'trash')
            )
            for prefix, base, icon in items:
                endpoint = types.new_class(f'{prefix}{model.__name__}', (base,), {})
                endpoint.check_permission = lambda self: (
                    cls().instantiate(self.request, self).check_permission()
                )
                endpoint.Meta = type(
                    'Meta', (), dict(
                        icon=icon, modal=prefix != 'Visualizar',
                        verbose_name=f'{prefix} {model._meta.verbose_name}'
                    )
                )
            if 'Meta' not in attrs:
                cls.Meta = type(
                    'Meta', (), dict(
                        icon=getattr(model._meta, 'icon', None), modal=False,
                        verbose_name=f'{model._meta.verbose_name_plural}'
                    )
                )
        return cls


class Endpoint(metaclass=EnpointMetaclass):
    cache = cache

    def __init__(self):
        self.request = None
        self.source = None
        self.instantiator = None
        self.cleaned_data = {}
        super().__init__()

    def configure(self, source, instantiator=None):
        self.source = source
        self.instantiator = instantiator
        return self

    def contextualize(self, request):
        self.request = request
        return self

    def objects(self, model):
        return apps.get_model(model).objects
        
    def get(self):
        return {}
    
    def post(self):
        return Response(message='Ação realizada com sucesso')
    
    def check_permission(self):
        return self.request.user.is_superuser
    
    def check_role(self, *names, superuser=True):
        if self.request.user.is_superuser and superuser:
            return True
        for name in names:
            if self.objects('slth.role').filter(username=self.request.user.username, name=name).exists():
                return True
        return False
    
    def redirect(self, url):
        raise JsonResponseException(dict(type='redirect', url=url))
    
    def getform(self, form):
        return form
    
    def process(self):
        data = self.get()
        title = self.get_verbose_name()
        if isinstance(data, models.QuerySet):
            data = data.contextualize(self.request).settitle(title)
        elif isinstance(data, Serializer):
            data = data.contextualize(self.request).settitle(title)
        elif isinstance(data, FormFactory):
            form = self.getform(data.settitle(title).form(self))
            if self.request.method == 'POST' or self.request.GET.get('form') == title:
                try:
                    self.cleaned_data = form.submit()
                    return self.post()
                except ValidationError as e:
                    raise JsonResponseException(dict(type='error', text='\n'.join(e.messages), errors={}))
            else:
                data = form
        elif isinstance(data, Form) or isinstance(data, ModelForm):
            data = data.settitle(title)
        return data
    
    def serialize(self):
        return serialize(self.process())
    
    def to_response(self):
        return ApiResponse(self.serialize(), safe=False)
    
    def formfactory(self, instance=None, method='POST') -> FormFactory:
        return FormFactory(instance, method=method)
    
    def serializer(self, instance=None) -> Serializer:
        return Serializer(instance or self).contextualize(self.request)
    
    @classmethod
    def is_child(cls):
        return False
    
    @classmethod
    def get_api_name(cls):
        return cls.__name__.lower()
    
    @classmethod
    def get_api_url(cls, arg=None):
        if arg:
            return f'/api/{cls.get_api_name()}/{arg}/'  
        return f'/api/{cls.get_api_name()}/'    
    
    @classmethod
    def get_pretty_name(cls):
        name = []
        for c in cls.__name__:
            if name and c.isupper():
                name.append(' ')
            name.append(c)
        return ''.join(name)
    
    @classmethod
    def get_qualified_name(cls):
        return '{}.{}'.format(cls.__module__, cls.__name__).lower()
    
    @classmethod
    def instantiate(cls, request, source):
        args = ()
        if cls.is_child():
            args = (source,) if cls.has_args() else ()
        else:
            args = (source.pk,) if cls.has_args() else ()
        return cls(*args).configure(source).contextualize(request)
    
    @classmethod
    def has_args(cls):
        return len(inspect.getfullargspec(cls.__init__).args) > 1
    
    @classmethod
    def get_api_url_pattern(cls):
        args = inspect.getfullargspec(cls.__init__).args[1:]
        pattern = '{}/'.format(cls.get_api_name())
        for arg in args:
            pattern = '{}{}/'.format(pattern, '<int:{}>'.format(arg))
        return pattern
    
    @classmethod
    def get_api_metadata(cls, request, base_url, pk=None):
        action_name = cls.get_metadata('verbose_name')
        icon = cls.get_metadata('icon')
        modal = cls.get_metadata('modal')
        if cls.is_child():
            url = append_url(base_url, f'action={cls.get_api_name()}')
            url = f'{url}&id={pk}' if pk else url
        else:
            url = build_url(request, f'/api/{cls.get_api_name()}/')
            url = f'{url}{pk}/' if pk else url
        return dict(type='action', title=action_name, name=action_name, url=url, key=cls.get_api_name(), icon=icon, modal=modal)
    
    @classmethod
    def get_metadata(cls, key, default=None):
        value = None
        metaclass = getattr(cls, 'Meta', None)
        if metaclass:
            value = getattr(metaclass, key, None)
        if value is None:
            if key == 'verbose_name':
                value = cls.get_pretty_name()
            if key == 'modal':
                value = issubclass(cls, EditEndpoint) or issubclass(cls, DeleteEndpoint) or issubclass(cls, Endpoint) or issubclass(cls, ChildEndpoint)
        return default if value is None else value
    
    def get_verbose_name(self):
        return self.get_metadata('verbose_name')

class PublicEndpoint(Endpoint):
    def check_permission(self):
        return True

class ModelEndpoint(Endpoint):
    def __init__(self):
        self.model = self.__orig_bases__[0].__args__[0]
        if isinstance(self.model, str):
            self.model = self.objects(self.model).get(pk=self.pk)
        super().__init__()

class AdminEndpoint(Generic[T], ModelEndpoint):

    def get(self) -> QuerySet:
        actions = [f'{prefix}{self.model.__name__.lower()}' for prefix in ('cadastrar', 'visualizar', 'editar', 'excluir')]
        return self.model.objects.all().actions(*actions)
    
class ListEndpoint(Generic[T], ModelEndpoint):
    def get(self) -> QuerySet:
        return self.model.objects

class AddEndpoint(Generic[T], ModelEndpoint):
    def __init__(self):
        super().__init__()
        self.instance = self.model()

    def get(self) -> FormFactory:
        return self.model().formfactory()
    
    def get_instance(self):
        return self.instance

class ModelInstanceEndpoint(ModelEndpoint):
    def __init__(self, pk):
        super().__init__()
        self.instance = self.model.objects.get(pk=pk)

    def get_instance(self):
        return self.instance

class InstanceEndpoint(Generic[T], ModelInstanceEndpoint):

    def formfactory(self) -> FormFactory:
        return FormFactory(self.get_instance())

    def serializer(self) -> Serializer:
        return Serializer(self.get_instance()).contextualize(self.request)
    

class ViewEndpoint(Generic[T], ModelInstanceEndpoint):

    class Meta:
        icon = 'eye'
        modal = False
        verbose_name = 'Visualizar'

    def get(self) -> Serializer:
        return self.get_instance().serializer().contextualize(self.request)
    
    def get_verbose_name(self):
        return str(self.get_instance())
        
class EditEndpoint(Generic[T], ModelInstanceEndpoint):
    def get(self) -> FormFactory:
        return self.get_instance().formfactory()
    
class DeleteEndpoint(Generic[T], ModelInstanceEndpoint):
    def get(self) -> FormFactory:
        return self.formfactory(self.get_instance()).fields()
    
    def post(self):
        self.get_instance().delete()
        return super().post()

    
class ChildEndpoint(Endpoint):

    @classmethod
    def is_child(cls):
        return True
    
    def check_permission(self):
        return True

class Add(ChildEndpoint):
    class Meta:
        icon = 'plus'
        verbose_name = 'Cadastrar'

    def get(self) -> FormFactory:
        return self.source.model().formfactory()
    
class ChildInstanceEndpoint(ChildEndpoint):
    def __init__(self, instance):
        self.instance = instance
        super().__init__()

    def get_instance(self):
        return self.instance
    
    def formfactory(self) -> FormFactory:
        return FormFactory(self.get_instance())
    
    def serializer(self) -> Serializer:
        return Serializer(self.get_instance()).contextualize(self.request)
    
class View(ChildInstanceEndpoint):
    class Meta:
        icon = 'eye'
        modal = False
        verbose_name = 'Visualizar'
    
    def get(self) -> Serializer:
        return self.get_instance().serializer()

class Edit(ChildInstanceEndpoint):
    class Meta:
        icon = 'pen'
        verbose_name = 'Editar'
    def get(self) -> FormFactory:
        return self.get_instance().formfactory()
   
class Delete(ChildInstanceEndpoint):
    class Meta:
        icon = 'trash'
        verbose_name = 'Excluir'
    
    def get(self):
        return self.formfactory().fields()
    
    def post(self):
        self.get_instance().delete()
        return super().post()

class Login(PublicEndpoint):
    username = forms.CharField(label='Username')
    password = forms.CharField(label='Senha')

    class Meta:
        modal = False
        icon = 'sign-in'
        verbose_name = 'Login'

    def get(self):
        code = self.request.GET.get('code')
        if code:
            user = oauth.authenticate(code)
            if user:
                token = Token.objects.create(user=user)
                return Response(message='Bem-vindo!', redirect='/api/dashboard/', store=dict(token=token.key, application=None))
        return self.formfactory().fields('username', 'password')
    
    def post(self):
        user = authenticate(self.request, username=self.cleaned_data.get('username'), password=self.cleaned_data.get('password'))
        if user:
            token = Token.objects.create(user=user)
            return Response(message='Bem-vindo!', redirect='/api/dashboard/', store=dict(token=token.key, application=None))
        else:
            raise ValidationError('Login e senha não conferem')

class Logout(Endpoint):
    class Meta:
        modal = False
        verbose_name = 'Sair'
        
    def get(self):
        return Response(message='Logout realizado com sucesso.', redirect='/api/home/', store=dict(token=None, application=None))

    def check_permission(self):
        return self.request.user.is_authenticated

class Icons(PublicEndpoint):
    class Meta:
        modal = True
        verbose_name = 'Icons'

    def get(self):
        return IconSet()


class Search(Endpoint):
    def get(self):
        key = '_options_'
        options = self.cache.get(key, [])
        term = self.request.GET.get('term')
        if options is None and APPLICATON['dashboard']['search']:
            options = []
            for endpoint in APPLICATON['dashboard']['search']:
                cls = ENDPOINTS[endpoint]
                url = build_url(self.request, cls.get_api_url())
                verbose_name = cls.get_metadata('verbose_name')
                options.append(dict(id=url, value=verbose_name))
            self.cache.set(key, options)
        if term:
            result = []
            for option in options:
                if slugify(term.lower()) in slugify(option['value'].lower()):
                    result.append(option)
        else:
            result = options
        return result[0:10]
    
class Users(ListEndpoint[User]):

    class Meta:
        modal = False
        verbose_name = 'Usuários'

    def get(self):
        return (
            super().get().search('username', 'email')
            .filters('is_superuser', 'is_active')
            .fields('username', 'email', 'is_superuser', 'get_roles')
            .actions('add', 'view', 'edit', 'delete', 'sendpushnotification', 'changepassword')
        )
    
class ChangePassword(ChildInstanceEndpoint):
    password = forms.CharField(label='Senha', required=False)

    class Meta:
        icon = 'user-lock'
        verbose_name = 'Alterar Senha'

    def get(self):
        return self.formfactory().fields('password')

    def post(self):
        self.instance.set_password(self.cleaned_data['password'])
        self.instance.save()
        return super().post()

class Home(PublicEndpoint):
    class Meta:
        verbose_name = ''

    def get(self):
        cls = ENDPOINTS[APPLICATON['index']]
        self.redirect(cls.get_api_url())


class Dashboard(Endpoint):
    class Meta:
        verbose_name = ''
        
    def get(self):
        if self.request.user.is_authenticated:
            serializer = Serializer(request=self.request)
            if APPLICATON['dashboard']['boxes']:
                boxes = Boxes('Acesso Rápido')
                for endpoint in APPLICATON['dashboard']['boxes']:
                    cls = ENDPOINTS[endpoint]
                    if cls().contextualize(self.request).check_permission():
                        icon = cls.get_metadata('icon', 'check')
                        label = cls.get_metadata('verbose_name')
                        url = build_url(self.request, cls.get_api_url())
                        boxes.append(icon, label, url)
                serializer.append('Acesso Rápido', boxes)
            if APPLICATON['dashboard']['top']:
                group = serializer.group('Top')
                for endpoint in APPLICATON['dashboard']['top']:
                    cls = ENDPOINTS[endpoint]
                    if cls.instantiate(self.request, self.request.user).check_permission():
                        group.endpoint(cls.get_metadata('verbose_name'), cls, wrap=False)
                group.parent()
            if APPLICATON['dashboard']['center']:
                for endpoint in APPLICATON['dashboard']['center']:
                    cls = ENDPOINTS[endpoint]
                    serializer.endpoint(cls.get_metadata('verbose_name'), cls, wrap=False)
            return serializer
        else:
            self.redirect('/api/login/')

    def check_permission(self):
        return self.request.user.is_authenticated
    
class Application(PublicEndpoint):
    def get(self):
        user = None
        navbar = None
        menu = None
        icon = build_url(self.request, APPLICATON['logo'])
        logo = build_url(self.request, APPLICATON['logo'])
        if self.request.user.is_authenticated:
            user = self.request.user.username.split()[0].split('@')[0]
        navbar = Navbar(
            title=APPLICATON['title'], subtitle=APPLICATON['subtitle'],
            logo=logo, user=user
        )
        for entrypoint in ['actions', 'usermenu', 'adder', 'settings', 'tools']:
            if APPLICATON['dashboard'][entrypoint]:
                for endpoint in APPLICATON['dashboard'][entrypoint]:
                    cls = ENDPOINTS[endpoint]
                    if cls().instantiate(self.request, self).check_permission():
                        label = cls.get_metadata('verbose_name')
                        url = build_url(self.request, cls.get_api_url())
                        modal = cls.get_metadata('modal', False)
                        icon = cls.get_metadata('icon', None)
                        navbar.add_action(entrypoint, label, url, modal, icon=icon)
        if APPLICATON['menu']:
            items = []
            def get_item(k, v):
                if isinstance(v, dict):
                    icon, label = k.split(':') if ':' in k else (None, k)
                    subitems = []
                    for k1, v1 in v.items():
                        subitem = get_item(k1, v1)
                        if subitem:
                            subitems.append(subitem)
                    if subitems:
                        return dict(dict(icon=icon, label=label, items=subitems))
                else:
                    cls = ENDPOINTS.get(v)
                    if cls:
                        if cls().instantiate(self.request, self).check_permission():
                            icon, label = k.split(':') if ':' in k else (None, k)
                            url = build_url(self.request, cls.get_api_url())
                            return dict(dict(label=label, url=url, icon=icon))
                    else:
                        print(v)
            for k, v in APPLICATON['menu'].items():
                item = get_item(k, v)
                if item:
                    items.append(item)
            profile = Profile.objects.filter(user=self.request.user).first() if user else None
            photo_url = profile.photo.url if profile and profile.photo else '/static/images/user.svg'
            menu = Menu(items, user=user, image=build_url(self.request, photo_url))

        footer = Footer(APPLICATON['version'])
        return Application_(icon=icon, navbar=navbar, menu=menu, footer=footer, oauth=oauth.providers())
    
class Manifest(PublicEndpoint):

    class Meta:
        verbose_name = 'Manifest'

    def get(self):
        return dict(
            {
                "name": APPLICATON['title'],
                "short_name": APPLICATON['title'],
                "lang": 'pt-BR',
                "start_url": build_url(self.request, "/app/dashboard/"),
                "scope": build_url(self.request, "/app/"),
                "display": "standalone",
                "icons": [{
                    "src": build_url(self.request, APPLICATON['logo']),
                    "sizes": "192x192",
                    "type": "image/png"
                }]
            }
        )

    def check_permission(self):
        return True


class PushSubscribe(Endpoint):

    class Meta:
        verbose_name = 'Subscrever para Notificações'

    def post(self):
        data = json.loads(self.request.POST.get('subscription'))
        device = self.request.META.get('HTTP_USER_AGENT', '')
        qs = PushSubscription.objects.filter(user=User.objects.first(), device=device)
        qs.update(data=data) if qs.exists() else PushSubscription.objects.create(
            user=User.objects.first(), data=data, device=device
        )
        return Response()

    def check_permission(self):
        return True
    
class PushSubscriptions(ListEndpoint[PushSubscription]):
    class Meta:
        verbose_name = 'Notificações'

class SendPushNotification(ChildInstanceEndpoint):
    message = forms.CharField(label='Texto', widget=forms.Textarea())
    class Meta:
        icon = 'mail-bulk'
        verbose_name = 'Enviar Notificação'

    def get(self):
        return self.formfactory().fields('message')

    def post(self):
        send_push_web_notification(self.instance, self.cleaned_data['message'])
        return Response(message='Notificação enviada com sucesso.')

class EditProfile(Endpoint):
    password = forms.CharField(label='Senha', required=False)
    password2 = forms.CharField(label='Confirmação', required=False)

    class Meta:
        verbose_name = 'Editar Perfil'

    def __init__(self, *args, **kwargs):
        self.instance = None
        super().__init__(*args, **kwargs)
       
    def get(self):
        self.instance = Profile.objects.filter(user=self.request.user).first() or Profile(user=self.request.user)
        return (
            self.formfactory(self.instance)
            .fieldset('Dados Gerais', ('photo',))
            .fieldset('Dados de Acesso', (('password', 'password2'),))
        )
    
    def post(self):
        self.instance.save()
        if self.cleaned_data.get('password'):
            self.instance.user.set_password(self.cleaned_data.get('password'))
            self.instance.user.save()
        return Response(message='Ação realizada com sucesso.', redirect='/api/dashboard/', store=dict(application=None))

    def check_permission(self):
        return self.request.user.is_authenticated




import datetime
import unicodedata
import os
import sys
import time

from django.conf import settings
from selenium import webdriver
from selenium.common.exceptions import (
    ElementNotInteractableException,
    StaleElementReferenceException,
    WebDriverException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.support.ui import Select


from selenium.webdriver import Remote, Firefox
from selenium.webdriver.remote.file_detector import LocalFileDetector
from selenium.webdriver.remote.webelement import WebElement

REMOTE = 'SELENIUM_HOST' in os.environ

def to_label_case(text):
    return unicodedata.normalize('NFD', text.replace('-', '').replace('_', '').lower())

def _upload(self, filename):
    return filename

WebElement._upload = _upload

class Browser(Remote if REMOTE else Firefox):
    def __init__(self, server_url, options=None, verbose=True, slowly=False, maximize=True, headless=True):
        if not options:
            options = Options()
        if maximize:
            options.add_argument("--start-maximized")
        else:
            options.add_argument("--window-size=720x800")
        if headless and "-v" not in sys.argv:
            options.add_argument("--headless")
        
        if REMOTE:
            url = 'http://{}:{}'.format(
                os.environ.get('SELENIUM_HOST', '127.0.0.1'),
                os.environ.get('SELENIUM_POST', '4444')
            )
            super().__init__(command_executor=url, options=options)
            self.file_detector = LocalFileDetector()
        else:
             super().__init__(options=options)

        self.cursor = None
        self.verbose = verbose
        self.slowly = slowly
        self.server_url = server_url
        self.headless = headless

        if maximize:
            self.maximize_window()
        else:
            self.set_window_position(700, 0)
            self.set_window_size(720, 800)
        self.switch_to.window(self.current_window_handle)

    def slow(self, slowly=True):
        self.slowly = slowly

    def is_cursor_inside(self, tag_name):
        if self.cursor:
            if self.cursor.tag_name == tag_name:
                self.debug(f"Cursor {self.cursor.tag_name} is inside {tag_name}")
                return True
            elif self.cursor.tag_name == "html":
                self.debug(f"Cursor {self.cursor.tag_name} is NOT inside {tag_name}")
                return False
            else:
                parent = self.cursor.find_element("xpath", "..")
                while True:
                    self.debug(f"Checking {parent.tag_name}")
                    if parent.tag_name == tag_name:
                        self.debug(f"Cursor {self.cursor.tag_name} is inside {tag_name}")
                        return True
                    elif parent.tag_name == "html":
                        return False
                    else:
                        parent = parent.find_element("xpath", "..")
        self.debug(f"Cursor {self.cursor.tag_name} is NOT inside {tag_name}")
        return False

    def initialize_cursor(self):
        try:
            dialogs = super().find_elements(By.TAG_NAME, "dialog")
            if dialogs:
                if self.cursor is None or not self.is_cursor_inside("dialog"):
                    self.cursor = dialogs[0]
            else:
                if self.cursor is None or not self.is_cursor_inside("html"):
                    self.cursor = super().find_element(By.TAG_NAME, "html")
            self.cursor.tag_name
        except StaleElementReferenceException:
            self.cursor = super().find_element(By.TAG_NAME, "html")
        self.debug(f"Cursor is at tag {self.cursor.tag_name}")

    def find_elements(self, by, value):
        self.initialize_cursor()
        while True:
            elements = self.cursor.find_elements(by, value)
            if elements:
                self.debug(f"Element {value} found in {self.cursor.tag_name}.")
                return elements
            elif self.cursor.tag_name == "dialog":
                self.debug("Recursive search stopped inside dialog!")
                break
            if self.cursor.tag_name != "html":
                self.debug(f"Element {value} not found in {self.cursor.tag_name}. Searching in the parent element...")
                self.cursor = self.cursor.find_element("xpath", "..")
            else:
                raise WebDriverException(f"Element {value} NOT FOUND in {self.cursor.tag_name}.")
        return []

    def find_element(self, by, value):
        self.initialize_cursor()
        while True:
            elements = self.cursor.find_elements(by, value)
            if elements:
                self.debug(f"Element {value} found in {self.cursor.tag_name}.")
                return elements[0]
            elif self.cursor.tag_name == "dialog":
                self.debug("Recursive search stopped inside dialog!")
                raise WebDriverException(f"Element {value} not found in the dialog.")
            if self.cursor.tag_name != "html":
                self.debug(f"Element {value} not found in {self.cursor.tag_name}. Searching in the parent element...")
                self.cursor = self.cursor.find_element("xpath", "..")
            else:
                raise WebDriverException(f"Element {value} NOT FOUND in {self.cursor.tag_name}.")

    def wait(self, seconds=2):
        time.sleep(seconds)

    def watch(self, e):
        self.save_screenshot("/tmp/test.png")
        if self.headless:
            raise e
        else:
            breakpoint()
            raise e

    def print(self, message):
        if self.verbose:
            print(message)

    def debug(self, message):
        # print(message)
        pass

    def execute_script(self, script, *args):
        super().execute_script(script, *args)
        if self.slowly:
            self.wait(3)

    def open(self, url):
        if url.startswith("http"):
            self.get(url.replace("http://localhost:8000", self.server_url))
        else:
            self.get(f"{self.server_url}{url}")

    def reload(self):
        self.open(self.current_url)

    def enter(self, name, value, submit=False, count=4):
        if callable(value):
            value = value()
        if type(value) == datetime.date:
            value = value.strftime("%Y-%d-%m")
        self.print('{} "{}" for "{}"'.format("Entering", value, name))
        if value:
            value = str(value)
            if len(value) == 10 and value[2] == "/" and value[5] == "/":
                value = datetime.datetime.strptime(value, "%d/%m/%Y").strftime("%Y-%m-%d")
        try:
            widget = self.find_element("css selector", f'[data-label="{to_label_case(name)}"]')
            if widget.tag_name == "input" and widget.get_property("type") == "color":
                self.execute_script(
                    f'document.querySelector(\'input[data-label="{to_label_case(name)}"]\').value = "{value}";'
                )
            else:
                if widget.tag_name == "input" and widget.get_property("type") == "file":
                    value = os.path.join(settings.BASE_DIR, value)
                widget.clear()
                widget.send_keys(value)
        except WebDriverException as e:
            if count:
                self.wait()
                self.enter(name, value, submit, count - 1)
            else:
                self.watch(e)
        if self.slowly:
            self.wait(2)

    def choose(self, name, value, count=4):
        self.print('{} "{}" for "{}"'.format("Choosing", value, name))
        try:
            widgets = self.find_elements(By.CSS_SELECTOR, f'[data-label="{to_label_case(name)}"]')
            if widgets:
                if widgets[0].tag_name.lower() == "select":
                    select = Select(widgets[0])
                    select.select_by_visible_text(value)
                elif widgets[0].tag_name.lower() == "input":
                    widgets[0].send_keys(value)
                    for i in range(0, 10):
                        print('Trying ({}) click at "{}"...'.format(i, value))
                        self.wait(0.5)
                        try:
                            super().find_element(By.CSS_SELECTOR, f'.autocomplete-item[data-label*="{to_label_case(value)}"]').click()
                            self.wait(0.5)
                            break
                        except WebDriverException:
                            pass
                elif widgets[0].get_dom_attribute("type") == "radio":
                    widgets[0].click()
                elif widgets[0].get_dom_attribute("type") == "checkbox":
                    widgets[0].click()
            else:
                raise WebDriverException()
        except WebDriverException as e:
            if count:
                self.wait()
                self.choose(name, value, count - 1)
            else:
                self.watch(e)
        if self.slowly:
            self.wait(2)

    def see(self, text, flag=True, count=4):
        if flag:
            self.print(f'See "{text}"')
            try:
                assert text in self.find_element(By.TAG_NAME, "body").text
            except AssertionError as e:
                if count:
                    self.wait()
                    self.see(text, flag, count - 1)
                else:
                    self.watch(e)
            if self.slowly:
                self.wait(2)
        else:
            self.print(f'Can\'t see "{text}"')
            try:
                assert text not in self.find_element(By.TAG_NAME, "body").text
            except AssertionError as e:
                if count:
                    self.wait()
                    self.see(text, flag, count - 1)
                else:
                    self.watch(e)
            if self.slowly:
                self.wait(2)

    def look_at(self, text, count=4):
        self.print(f'Loking at "{text}"')
        try:
            self.cursor = self.find_element(By.CSS_SELECTOR, f'[data-label="{to_label_case(text)}"]')
            if self.cursor:
                self.execute_script("arguments[0].scrollIntoView();", self.cursor)
                self.debug(f"Cursor is now at {self.cursor.tag_name}")
        except WebDriverException as e:
            if count:
                self.wait()
                self.look_at(text, count - 1)
            else:
                self.watch(e)
        if self.slowly:
            self.wait(2)

    def click(self, text, count=4):
        self.print(f'Clicking "{text}"')
        try:
            elements = self.find_elements("css selector", f'[data-label="{to_label_case(text)}"]')
            if elements:
                for element in elements:
                    try:
                        element.click()
                        break
                    except ElementNotInteractableException:
                        pass
            else:
                raise WebDriverException()
        except WebDriverException as e:
            if count:
                self.wait()
                self.click(text, count=count - 1)
            else:
                self.watch(e)

    def close(self, seconds=0):
        self.wait(seconds)
        super().close()



from django.core.cache import cache
from django.core.management.commands.test import Command


class Command(Command):
    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("--from", nargs="+", help="Run test from a specific step")
        parser.add_argument("--save", action="store_true", help="Save the steps in .steps directory")
        parser.add_argument("--nobuild", action="store_true", help="Does not build the frontend application")
        parser.add_argument("--restore", nargs="+", help="Restore development database from a specific step")
        parser.add_argument("--browser", action="store_true", help="Run test in the browser in fast speed")
        parser.add_argument("--tutorial", action="store_true", help="Run test in the browser in slow speed")
        parser.add_argument("--write", action="store_true", help="Log browser actions in the terminal")

    def handle(self, *args, **options):
        from_step = int(options.get("from")[0]) if options.get("from") else 0
        restore = int(options.get("restore")[0]) if options.get("restore") else None
        if from_step > 1:
            cache.set("FROM_STEP", from_step - 1)
        cache.set("SAVE_STEP", options["save"])
        cache.set("NOBUILD", options["nobuild"])
        cache.set("RESTORE", restore)
        cache.set("EXPLAIN", options["tutorial"])
        cache.set("LOG_ACTION", 1 if options["write"] else 0)
        cache.set("HEADLESS", not options["browser"] and not options["tutorial"] and not options["write"])
        super().handle(*args, **options)



# Generated by Django 4.1 on 2024-03-16 23:08

from django.conf import settings
from django.db import migrations
from slth.db import models
import slth


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('slth', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Email',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('from_email', models.EmailField(max_length=254, verbose_name='Remetente')),
                ('to', models.TextField(help_text='Separar endereços de e-mail por ",".', verbose_name='Destinatário')),
                ('subject', models.CharField(max_length=255, verbose_name='Assunto')),
                ('content', models.TextField(verbose_name='Conteúdo')),
                ('sent_at', models.DateTimeField(null=True, verbose_name='Data/Hora')),
            ],
            options={
                'verbose_name': 'E-mail',
                'verbose_name_plural': 'E-mails',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('username', models.CharField(db_index=True, max_length=50)),
                ('name', models.CharField(db_index=True, max_length=50)),
                ('scope', models.CharField(db_index=True, max_length=50, null=True)),
                ('model', models.CharField(db_index=True, max_length=50, null=True)),
                ('value', models.IntegerField(db_index=True, null=True, verbose_name='Value')),
                ('active', models.BooleanField(default=True, null=True, verbose_name='Active')),
            ],
            options={
                'verbose_name': 'Papel',
                'verbose_name_plural': 'Papéis',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PushSubscription',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('device', models.CharField(max_length=255, verbose_name='Dispositivo')),
                ('data', models.JSONField(verbose_name='Dados da Inscrição')),
                ('user', models.ForeignKey(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='Usuário')),
            ],
            options={
                'verbose_name': 'Inscrição de Notificação',
                'verbose_name_plural': 'Inscrições de Notificação',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Error',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(verbose_name='Data/Hora')),
                ('traceback', models.TextField(verbose_name='Rastreamento')),
                ('user', models.OneToOneField(null=True, on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='Usuário')),
            ],
            options={
                'verbose_name': 'Erro',
                'verbose_name_plural': 'Erros',
            },
            bases=(models.Model,),
        ),
    ]



# Generated by Django 4.1 on 2024-03-10 00:01

from django.conf import settings
from django.db import migrations
from slth.db import models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Token',
            fields=[
                ('key', models.CharField(max_length=40, primary_key=True, serialize=False, verbose_name='Key')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='Created')),
                ('user', models.ForeignKey(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Token',
                'verbose_name_plural': 'Tokens',
            },
        ),
        migrations.CreateModel(
            name='Foto',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('photo', models.ImageField(upload_to='', verbose_name='Foto')),
                ('user', models.OneToOneField(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Foto',
                'verbose_name_plural': 'Fotos',
            },
        ),
    ]



from pywebpush import webpush

def send_push_web_notification(user, text):
    for subscription in user.pushsubscription_set.all():
        data = webpush(
            subscription_info=subscription.data,
            data=text,
            vapid_private_key="GoFJpuTAdhepzfxOHdrW7u2ONh7V8ZIjPkjgpWSS3ks",
            vapid_claims={"sub": "mailto:admin@admin.com"}
        )
        print(data)
